/*
 * Copyright 2010 Portland Webworks, Inc.
 *
 * This file is part of PWW Commons.
 *
 * PWW Commons is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 * Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * PWW Commons is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with PWW Commons.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

package com.portlandwebworks.commons.dao;

import com.portlandwebworks.commons.domain.IEntity;

import java.io.Serializable;
import java.util.Collection;
import java.util.List;

/**
 * This software is the property of Portland Webworks.
 *
 * This interface defines an entity-specific means of 
 *
 * @Author rnorthup
 */
public interface IGenericDaoBase<T extends IEntity>
{
	/**
	 * Retrieves an entity by its unique identifier.
	 *
	 * @param id
	 * @throws com.portlandwebworks.commons.dao.DataNotFoundException if an entity of the type requested does not exist with the
	 * 			identifier provided.
	 * @return
	 */
	public T findById(Serializable id) throws DataNotFoundException;

	/**
	 * Retrieves an entity by its unique identifier, eagerly fetching those fields specified.
	 *
	 * @param id
	 * @param fetchOptions
	 * @throws com.portlandwebworks.commons.dao.DataNotFoundException if an entity of the type requested does not exist with the
	 * 			identifier provided.
	 * @return
	 */
	public T findById(Serializable id, Enum... fetchOptions) throws DataNotFoundException;

	/**
	 *
	 * @return
	 */
	public List<T> findAll();

	/**
	 * Retrieves a collection of entities by their primary key values.
	 * 
	 * @param ids
	 * @return
	 * @throws DataNotFoundException
	 */
	public List<T> findAllByIds(Serializable[] ids) throws DataNotFoundException;

	/**
	 * Generic method to return a collection of entities that match the criteria provided.
	 * @param criteria
	 * @return
	 */
	public List<T> findAllByCriteria(SearchCriteria criteria);

	/**
	 * Generic method to return a "page" of search results suitable for display in a grid view.
	 *
	 * @param criteria
	 * @param options
	 * @return
	 */
	public GridDisplayResults<T> search(SearchCriteria criteria, GridDisplayOptions options);

	/**
	 * Inserts a new instance of entity into the data source.
	 *
	 * @param entity
	 * @return
	 */
	public Serializable insert(T entity);

	/**
	 * Added because Jacob doesn't like to write code.
	 * 
	 * @param entity
	 */
	public void saveOrUpdate(T entity);

	/**
	 * Updates an entity previously stored in the data source.
	 *
	 * @param entity
	 */
	public void update(T entity);

	/**
	 * Removes an entity from the data source.
	 *
	 * @param entity
	 */
	public void delete(T entity);

	/**
	 * Removes an entity from the data source that has not been loaded by referencing it by its identifier.
	 * @param id
	 */
	public void deleteById(Serializable id);

	/**
	 * Removes a collection of entities from the data source.
	 * @param entities
	 */
	public void deleteAll(Collection<T> entities);

	/**
	 * A generic way of retrieving a sequential value from the database that is guaranteed to be unique within the
	 * scope of the counter name provided.
	 * 
	 * @param counterName
	 * @return
	 */
	public Long getCounterValue(String counterName);

	/**
	 * This should force the DAO implementation to flush all pending changes to the database without yet committing.
	 */
	public void saveChanges();

}