/*
 * Copyright 2010 Portland Webworks, Inc.
 *
 * This file is part of PWW Commons.
 *
 * PWW Commons is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 * Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * PWW Commons is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with PWW Commons.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

package com.portlandwebworks.commons.hibernate;

import java.io.Serializable;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.sql.Types;
import java.util.Date;

import org.hibernate.Hibernate;
import org.hibernate.HibernateException;
import org.hibernate.usertype.UserType;

/**
* @author Bertrand Renuart
*
* Hibernate User Type to persist java.util.Date object as java.sql.Timestamp(s)
*
* The default behavior of Hibernate is to return the java.sql.Timestamp object
* retrieved from the underlying Resultset. Unfortunately, because of the extra
* precision (to the nanoseconds) of Timestamp, compare methods (including
* equals & compareTo) fail when used with java.util.Date instances.
*
* This user type will convert the java.sql.Timestamp back into java.util.Date
* instances (loosing the extra precision) before returning the data to the
* user.
*/
public class DateTimeType implements UserType
{
   private static final int[] SQL_TYPES = new int[] { Types.TIMESTAMP };


   //
   // ---- Constructors ---------------------------------------------------------
   //

   /**
    *
    */
   public DateTimeType()
   {
      super();
   }

   //
   // ---- UserType interface implementation ------------------------------------
   //

   /**
    * The class returned by this type factory.
    *
    * @see org.hibernate.usertype.UserType#returnedClass()
    */
   public Class returnedClass() {
      return java.util.Date.class;
   }

   /**
    * Return the SQL type codes for the columns mapped by this type
    *
    * @see org.hibernate.usertype.UserType#sqlTypes()
    */
   public int[] sqlTypes() {
      return SQL_TYPES;
   }

   /**
    * Compare two instances of the class mapped by this type for persistence "equality".
    *
    * @see org.hibernate.usertype.UserType#equals(java.lang.Object, java.lang.Object)
    */
   public boolean equals(Object x, Object y) throws HibernateException {
      if( x==y ) return true;
      if( (x==null) || (y == null) ) return false;

      Date xDate = (Date) x;
      Date yDate = (Date) y;

      return x.equals(y);
   }

	public int hashCode(Object o) throws HibernateException
	{
		Date date = (Date)o;
		return date.hashCode();
	}


	/**
    * Retrieve an instance of the mapped class from a JDBC resultset.
    *
    * @see org.hibernate.usertype.UserType#nullSafeGet(java.sql.ResultSet, java.lang.String[], java.lang.Object)
    */
   public Object nullSafeGet(ResultSet rs, String[] names, Object owner)
      throws HibernateException, SQLException
   {
       // extract Timestamp from the result set
      Timestamp timestamp = (Timestamp) Hibernate.TIMESTAMP.nullSafeGet(rs, names[0]);

      // return the value as a java.util.Date (dropping the nanoseconds)
      if( timestamp==null )
      {
         return null;
      }
      else
      {
         return new Date(timestamp.getTime());
      }
   }


   /**
    * Write an instance of the mapped class to a prepared statement
    *
    * @see org.hibernate.usertype.UserType#nullSafeSet(java.sql.PreparedStatement, java.lang.Object, int)
    */
   public void nullSafeSet(PreparedStatement st, Object value, int index)
      throws HibernateException, SQLException
   {
      // handle the NULL special case immediately
      if( value==null )
      {
         st.setTimestamp(index, null);
         return;
      }


      // make sure the received value is of the right type
      if( ! Date.class.isAssignableFrom(value.getClass()) ) {
         throw new IllegalArgumentException("Received value is not a [java.util.Date] but [" + value.getClass() + "]");
      }

      // set the value into the resultset
      Timestamp tstamp = null;
      if( (value instanceof Timestamp) )
      {
         tstamp = (Timestamp) value;
      }
      else
      {
         tstamp = new Timestamp( ((Date) value).getTime() );
      }

      st.setTimestamp(index, tstamp);
   }


   /**
    * Return a deep copy of the persistent state, stopping at entities and at collections
    *
    * @see org.hibernate.usertype.UserType#deepCopy(java.lang.Object)
    */
   public Object deepCopy(Object value) throws HibernateException {
      if( value==null )
      {
         return null;
      }
      else
      {
         return ((Date) value).clone();
      }
   }


   /**
    * Are objects of this type mutable?
    *
    * Date instances are mutable (although the date.set() methods are deprecated, there are
    * still accessible)
    *
    * @see org.hibernate.usertype.UserType#isMutable()
    */
   public boolean isMutable() {
      return true;
   }

	public Serializable disassemble(Object value) throws HibernateException
	{
		if (value == null)
			return null;

		return (Date)((Date)value).clone();
	}

	public Object assemble(Serializable cached, Object owner) throws HibernateException
	{
		Date date = (Date)cached;
		if (date == null) {
			return null;
		}
		return date.clone();
	}

	public Object replace(Object original, Object target, Object owner) throws HibernateException
	{
		Date date = (Date)original;
		return date.clone();
	}

}