/*
 * Copyright 2010 Portland Webworks, Inc.
 *
 * This file is part of PWW Commons.
 *
 * PWW Commons is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 * Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * PWW Commons is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with PWW Commons.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

package com.portlandwebworks.commons.hibernate;

import com.portlandwebworks.commons.dao.*;
import com.portlandwebworks.commons.domain.*;
import org.hibernate.*;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.transaction.annotation.Propagation;
import org.springframework.transaction.annotation.Transactional;

import java.io.Serializable;
import java.lang.InstantiationException;
import java.util.Collection;
import java.util.List;


/**
 * This software is the property of Portland Webworks.
 *
 * This class should be extended for a DAO object that supports data access for a specific domain object and the
 * <code>HibernateDaoBase</code> class does not provide the necessary functionality.
 *
 * <p>
 * All newly added DAO methods should include @Transactional(propagation = Propagation.MANDATORY) to be consistent with
 * the DAO methods that already exist on this base class.

 * </p>
 *
 * @Author rnorthup
 */
public class GenericHibernateDaoBase<T extends IEntity> implements IGenericDaoBase<T>
{
	private Class<T> entityClass;

	private T entityInstance;

	protected HibernateDaoBase daoBase;

	/**
	 *
	 * @param entityClass
	 */
	public GenericHibernateDaoBase(Class<T> entityClass)
	{
		this.entityClass = entityClass;

		this.daoBase = new HibernateDaoBase();

		try {
			entityInstance = entityClass.newInstance();
		}
		catch (InstantiationException e) {
			throw new RuntimeException("No default constructor for " + entityClass.getName());
		}
		catch (IllegalAccessException e) {
			throw new RuntimeException("Default constructor not accessible for " + entityClass.getName());
		}
	}

	/**
	 * The class of the type returned by this DAO.
	 *
	 * @return
	 */
	protected Class<T> getEntityClass()
	{
		return entityClass;
	}

	/**
	 * The name of the entity used in HQL statements.
	 *
	 * @return
	 */
	public String getEntityName()
	{
		return daoBase.getEntityName(entityInstance);
	}

	/**
	 * This session factory is used by the DAO for all database access and should be set by the container at runtime.
	 *
	 * @param sessionFactory
	 */
	@Autowired(required = true)
	public void setSessionFactory(SessionFactory sessionFactory)
	{
		this.daoBase.setSessionFactory(sessionFactory);
	}

	/**
	 * Provides DAO subclasses with the ability to access the underlying hibernate session.
	 * 
	 * @return
	 */
	protected Session getSession()
	{
		return daoBase.getSession();
	}

	@Transactional(propagation = Propagation.MANDATORY)
	public T findById(Serializable id) throws DataNotFoundException, DataIntegrityException
	{
		return daoBase.findById(entityClass, id);
	}

	@Transactional(propagation = Propagation.MANDATORY)
	public T findById(Serializable id, Enum... fetchOptions) throws DataNotFoundException
	{
		return daoBase.findById(entityClass, id, fetchOptions);
	}

	@Transactional(propagation = Propagation.MANDATORY)
	public List<T> findAll()
	{
		return daoBase.findAll(entityClass);
	}

	@Transactional(propagation = Propagation.MANDATORY)
	public List<T> findAllByIds(Serializable[] ids) throws DataNotFoundException
	{
		return daoBase.findAllByIds(entityClass, ids);
	}

	@Transactional(propagation = Propagation.MANDATORY)
	public List<T> findAllByCriteria(SearchCriteria searchCriteria)
	{
		return daoBase.findAllByCriteria(entityClass, searchCriteria);
	}

	@Transactional(propagation = Propagation.MANDATORY)
	public GridDisplayResults<T> search(SearchCriteria searchCriteria, GridDisplayOptions options)
	{
		return daoBase.search(entityClass, searchCriteria, options);
	}

	@Transactional(propagation = Propagation.MANDATORY)
	public Serializable insert(T entity)
	{
		return daoBase.insert(entity);
	}

	@Transactional(propagation = Propagation.MANDATORY)
	public void saveOrUpdate(T entity)
	{
		daoBase.saveOrUpdate(entity);
	}

	@Transactional(propagation = Propagation.MANDATORY)
	public void update(T entity)
	{
		daoBase.update(entity);
	}

	@Transactional(propagation = Propagation.MANDATORY)
	public void delete(T entity)
	{
		daoBase.delete(entity);
	}

	@Transactional(propagation = Propagation.MANDATORY)
	public void deleteById(Serializable id)
	{
		daoBase.deleteById(entityClass, id);
	}

	@Transactional(propagation = Propagation.MANDATORY)
	public void deleteAll(Collection<T> entities)
	{
		daoBase.deleteAll(entities);
	}

	@Transactional(propagation = Propagation.MANDATORY)
	public Long getCounterValue(String counterName)
	{
		return daoBase.getCounterValue(counterName);
	}

	@Transactional(propagation = Propagation.MANDATORY)
	public void saveChanges()
	{
		daoBase.saveChanges();
	}
}