/*
 * Copyright 2010 Portland Webworks, Inc.
 *
 * This file is part of PWW Commons.
 *
 * PWW Commons is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 * Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * PWW Commons is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with PWW Commons.  If not, see
 * <http://www.gnu.org/licenses/>.
 */
package com.portlandwebworks.commons.hibernate;

import com.portlandwebworks.commons.dao.*;
import com.portlandwebworks.commons.domain.IEntity;
import com.portlandwebworks.commons.reflection.FieldUtils;
import org.apache.commons.beanutils.PropertyUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;
import org.hibernate.*;
import org.hibernate.criterion.*;
import org.hibernate.engine.NamedSQLQueryDefinition;
import org.hibernate.engine.query.sql.NativeSQLQueryReturn;
import org.hibernate.engine.query.sql.NativeSQLQueryScalarReturn;
import org.hibernate.impl.SessionFactoryImpl;
import org.hibernate.type.Type;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.orm.hibernate3.SessionFactoryUtils;

import javax.persistence.EmbeddedId;
import java.io.Serializable;
import java.lang.reflect.Array;
import java.lang.reflect.Field;
import java.math.BigDecimal;
import java.text.MessageFormat;
import java.util.*;

/**
 * This software is the property of Portland Webworks.
 *
 * @Author rnorthup
 */
public class HibernateDaoBase implements IDaoBase
{
	private Logger logger = Logger.getLogger(this.getClass());

	private SessionFactory sessionFactory;

	private HashMap<Class<?>, List<String>> fieldMultiplicities = new HashMap<Class<?>, List<String>>();

	protected String entityName;

	private static Map<Class<?>, Type> typeMappings = new HashMap<Class<?>, Type>();

	static {
		typeMappings.put(Hibernate.STRING.getReturnedClass(), Hibernate.STRING);
		typeMappings.put(Hibernate.INTEGER.getReturnedClass(), Hibernate.INTEGER);
		typeMappings.put(Hibernate.DOUBLE.getReturnedClass(), Hibernate.DOUBLE);
		typeMappings.put(Hibernate.FLOAT.getReturnedClass(), Hibernate.FLOAT);
		typeMappings.put(Hibernate.SHORT.getReturnedClass(), Hibernate.SHORT);
		typeMappings.put(Hibernate.DATE.getReturnedClass(), Hibernate.DATE);
		typeMappings.put(Hibernate.LONG.getReturnedClass(), Hibernate.LONG);
		typeMappings.put(Hibernate.BOOLEAN.getReturnedClass(), Hibernate.BOOLEAN);
		typeMappings.put(Hibernate.BYTE.getReturnedClass(), Hibernate.BYTE);
	}

	protected static Type getTypeForClass(Class<?> clazz)
	{
		Type type = typeMappings.get(clazz);
		if (type != null) {
			return type;
		}
		return Hibernate.STRING;
	}

	/**
	 *
	 */
	public HibernateDaoBase()
	{
	}

	/**
	 * This session factory is used by the DAO for all database access and should be set by the container at runtime.
	 *
	 * @param sessionFactory
	 */
	@Autowired(required = true)
	public void setSessionFactory(SessionFactory sessionFactory)
	{
		this.sessionFactory = sessionFactory;
	}

	/**
	 * The name of the entity used in HQL statements.
	 *
	 * @return
	 */
	public String getEntityName(IEntity entityInstance)
	{
		if (entityName == null) {
			entityName = getSession().getEntityName(entityInstance);
		}
		return entityName;
	}


	protected boolean hasMultiValueField(Class entityClass)
	{
		if (! fieldMultiplicities.containsKey(entityClass)) {
			//process this class
			List<String> fieldNames = DaoUtils.getMultiValueFieldNames(entityClass);
			fieldMultiplicities.put(entityClass, fieldNames);
		}
		return fieldMultiplicities.get(entityClass).size() > 0;		
	}

	/**
	 * Determines if the property represented by the enumeration is backed by a Collection.
	 *
	 * @param entityClass
	 * @param fieldName
	 * @return
	 */
	protected <T extends IEntity> boolean isMultiValueField(Class<T> entityClass, String fieldName)
	{
		if (! fieldMultiplicities.containsKey(entityClass)) {
			//process this class
			List<String> fieldNames = DaoUtils.getMultiValueFieldNames(entityClass);
			fieldMultiplicities.put(entityClass, fieldNames);
		}

		return fieldMultiplicities.get(entityClass).contains(fieldName);
	}

	/**
	 * Utility method to retrieve the default schema configured for the underlying Hibernate session.
	 *
	 * @return	The string name of the schema, or <code>null</code> if no default schema is configured.
	 */
	protected String getDefaultSchemaName()
	{
		// This interface doesn't have the necessary api.  Need to downcast to the impl.
		SessionFactory factory = getSession().getSessionFactory();
		if (factory instanceof SessionFactoryImpl) {
			SessionFactoryImpl impl = (SessionFactoryImpl)factory;
			return impl.getSettings().getDefaultSchemaName();
		}
		return null;
	}

	/**
	 * Provides DAO subclasses with the ability to access the underlying hibernate session.
	 *
	 * @return
	 */
	protected Session getSession()
	{
		return SessionFactoryUtils.getSession(sessionFactory, true);
	}

	/**
	 * Helper method for reducing the list results of a query down to a single result.
	 *
	 * @param query
	 * @param <T>
	 * @return
	 */
	@SuppressWarnings("unchecked")
	protected <T> T getSingleResult(Criteria query) throws DataNotFoundException, DataIntegrityException
	{
		List result = query.list();
		switch (result.size()) {
			case 0:
				throw new DataNotFoundException();
			case 1:
				return (T)result.get(0);
			default:
				throw new DataIntegrityException("More than one result found for query (expected one)");
		}
	}

	@SuppressWarnings("unchecked")
	public <T extends IEntity> T findById(Class<T> entityClass, Serializable id) throws DataNotFoundException, DataIntegrityException
	{
		T result = (T)getSession().get(entityClass, id);
		if (result == null) {
			throw new DataNotFoundException(entityClass, id);
		}
		return result;
	}

	@SuppressWarnings("unchecked")
	public <T extends IEntity> T findById(Class<T> entityClass, Serializable id, Enum... fetchOptions) throws DataNotFoundException
	{
		Criteria criteria = getSession().createCriteria(entityClass);

		Enum prefetch = applyPrefetch(entityClass, criteria, fetchOptions);

		criteria.add(Restrictions.eq("id", id));
		criteria.setResultTransformer(Criteria.DISTINCT_ROOT_ENTITY);

		T entity = (T)criteria.uniqueResult();

		if (entity == null) {
			throw new DataNotFoundException(entityClass, id);
		}

		applyPostfetch(entityClass, prefetch, entity, fetchOptions);

		return entity;
	}

	@SuppressWarnings({"unchecked"})
	public <T extends IEntity> List<T> findAll(Class<T> entityClass)
	{
		Criteria query = getSession().createCriteria(entityClass);

		applySorting(entityClass, query, null);

		return query.list();
	}

	@SuppressWarnings("unchecked")
	public <T extends IEntity> List<T> findAllByIds(Class<T> entityClass, Serializable[] ids) throws DataNotFoundException
	{
		Criteria query = getSession().createCriteria(entityClass);

		query.add(Restrictions.in("id", ids));

		query.setResultTransformer(Criteria.DISTINCT_ROOT_ENTITY);

		List list = query.list();

		if (list.size() < ids.length) {
			// Returned entity list size is less than size of requested ID array.  This means that at least one ID does
			// not represent a persisted entity.  Report an error by identifying the first ID that doesn't match the
			// ID of any returned entity.
			for (Serializable id : ids) {
				boolean match = false;
				for (Object obj : list) {
					IEntity entity = (IEntity)obj;
					if (entity.getId().equals(id)) {
						match = true;
						break;
					}
				}
				if (! match) {
					throw new DataNotFoundException(entityClass, id);
				}
			}
		}
		return list;
	}

	@SuppressWarnings("unchecked")
	public <T extends IEntity> List<T> findAllByCriteria(Class<T> entityClass, SearchCriteria searchCriteria)
	{
		Criteria query = getSession().createCriteria(entityClass);

		applySearchCriteria(searchCriteria, query);

		applySorting(entityClass, query, null);

		// Applying predicates may cause joins to occur if a predicate spans entity relationships.
		query.setResultTransformer(Criteria.DISTINCT_ROOT_ENTITY);

		return query.list();
	}

	@SuppressWarnings("unchecked")
	public <T extends IEntity> List<T> findAllByPredicate(Class<T> entityClass, Predicate... predicates)
	{
		Criteria query = getSession().createCriteria(entityClass);

		applyPredicates(query, Arrays.asList(predicates));

		// Applying predicates may cause joins to occur if a predicate spans entity relationships.
		query.setResultTransformer(Criteria.DISTINCT_ROOT_ENTITY);

		applySorting(entityClass, query, null);

		return query.list();
	}

	@SuppressWarnings("unchecked")
	public <T extends IEntity> T findByPredicate(Class<T> entityClass, Predicate... predicates) throws DataNotFoundException
	{
		Criteria query = getSession().createCriteria(entityClass);

		applyPredicates(query, Arrays.asList(predicates));

		// Applying predicates may cause joins to occur if a predicate spans entity relationships.
		query.setResultTransformer(Criteria.DISTINCT_ROOT_ENTITY);

		List<T> results = query.list();
		switch (results.size()) {
			case 0:
				return null;
			case 1:
				return results.get(0);
			default:
				throw new DataIntegrityException("Multiple values found query, expected zero or one");
		}
	}

	protected <T extends IEntity> Enum applyPrefetch(Class<T> entityClass, Criteria criteria, Enum...fetchOptions)
	{
		// This will be the singular 1-M entity field that is pre-fetched.  Hibernate only allows
		// pre-fetching one aggregate field when retrieving an entity.
		Enum prefetch = null;

		// Iterate through the fetch options, eager joining all 1-1 relationships, an skipping over any
		// 1-M relationships after the first one has been eager joined.
		for (Enum fetchOption : fetchOptions) {

			// The name of the relationship to fetch is not the same as the name of the annotation.
			FetchOption fetch = fetchOption.getClass().getAnnotation(FetchOption.class);
			if (fetch != null) {
				String join = fetch.value();
				logger.debug("Eager joining relationship: " + join);
				criteria.setFetchMode(join, FetchMode.JOIN);
			}
			else {
				if (isMultiValueField(entityClass, fetchOption.toString()) && prefetch != null) {
					logger.debug("Skipping fetch option " + fetchOption + " as eager load field in query: can only eager load one 1-M field (" + prefetch + ")");
				}
				else {
					logger.debug("Eager joining collection: " + fetchOption);
					criteria.setFetchMode(fetchOption.toString(), FetchMode.JOIN);
					prefetch = fetchOption;
				}
			}
		}

		return prefetch;
	}

	/**
	 * Eager loads all multi-valued fields on the entity provided that are represented by the fetch options specified,
	 * excluding the field that was previously fetched by join.
	 *
	 * @param entityClass
	 * @param prefetch
	 * @param entity
	 * @param fetchOptions
	 * @param <T>
	 */
	protected <T extends IEntity> void applyPostfetch(Class<T> entityClass, Enum prefetch, T entity, Enum... fetchOptions)
	{
		String field = null;

		try {
			// Post-fetch all requested aggregate fields except for the one that was pre-fetched.  The name
			// of the enum is the same as the bean property.
			for (Enum fetchOption : fetchOptions) {
				if (isMultiValueField(entityClass, fetchOption.toString()) && fetchOption != prefetch) {
					field = fetchOption.toString();
					logger.debug("Eager loading collection after query: " + field);
					Object obj = PropertyUtils.getProperty(entity, field);
					if (obj instanceof Collection)
						((Collection)obj).size();
					else if (obj instanceof Map)
						((Map)obj).size();
				}
			}
		}
		catch (Exception e) {
			throw new RuntimeException("Unhandled exception fetching field " + field , e);
		}

	}

	/**
	 * Internal convenience method for apply predicates to a hibernate Criteria instance.
	 * 
	 * @param query         The query
	 * @param predicates    The predicates
	 */
	protected void applyPredicates(Criteria query, List<Predicate> predicates)
	{
		/**
		 * This is the set of all entities that have been joined in the search query, mapped to their aliases.
		 */
		HashMap<String, String> joinedEntities = new HashMap<String, String>();

		// This guarantees a unique alias name.
		int aliasCount = 0;

		for (Predicate predicate : predicates) {
			String property = predicate.getProperty();
			String operand = property;
			String joinPath = predicate.getJoinPath();

			if (joinPath != null) {
				String alias = joinedEntities.get(joinPath);
				if (alias == null) {
					String targetRelationship;
					int pos = StringUtils.lastIndexOf(joinPath, ".");
					if (pos > 0) {
						targetRelationship = joinPath.substring(pos + 1);
					}
					else {
						targetRelationship = joinPath;
					}
					alias = StringUtils.substring(targetRelationship, 0, 3) + "_" + aliasCount++;

					// Apply the join operation using our newly created alias.
					query.createAlias(joinPath, alias);

					joinedEntities.put(joinPath, alias);
				}
				operand = MessageFormat.format("{0}.{1}", alias, property);
			}

			// If there's a dot in this property, it's a property on a nested element.
/*
			int lastDot = property.lastIndexOf(".");
			if (lastDot != -1) {

				// This is the query path leading up to the nested entity.
				String joinPath = property.substring(0, lastDot);

				// This is the name of the property on the nested element that is to be evaluated in the query.
				property = property.substring(lastDot + 1);

				//TODO: this does not effectively handle joins more than one association in length, which should
				//TODO: check their "parts" to see if they have already been previously joined, and leverage them
				String alias = joinedEntities.get(joinPath);
				if (alias == null) {
					// Build an alias for the target entity of the search operation.
					String targetRelationship;
					int pos = StringUtils.lastIndexOf(joinPath, ".");
					if (pos > 0) {
						targetRelationship = joinPath.substring(pos + 1);
					}
					else {
						targetRelationship = joinPath;
					}
					alias = targetRelationship + "_" + aliasCount++;

					// Apply the join operation using our newly created alias.
					query.createAlias(joinPath, alias);

					joinedEntities.put(joinPath, alias);
				}

				// The new operand is alias.propertyName
				operand = MessageFormat.format("{0}.{1}", alias, property);
			}
*/

			applyPredicate(query, operand, predicate.getOperator(), predicate.getValue());
		}
	}

	/**
	 * Convenience method for applying query search criteria to a query by scanning an object for @SearchCriterion
	 * annotated fields with non-empty values, and converting them into Criteria API calls.
	 *
	 * @param searchCriteria
	 * @param query
	 */
	protected void applySearchCriteria(SearchCriteria searchCriteria, Criteria query)
	{
		if (searchCriteria == null) {
			throw new IllegalArgumentException("Search criteria is required");
		}

		if (logger.isDebugEnabled()) {
			//TODO:
		}

		/**
		 * This is the set of all entities that have been joined in the search query, mapped to their aliases.
		 */
	//	HashMap<String, String> joinedEntities = new HashMap<String, String>();

		List<Predicate> predicates = null;

		for (Field field : FieldUtils.getDeclaredFields(searchCriteria.getClass())) {

			// Check to see if there's an annotation on this field.
			SearchCriterion criterion = field.getAnnotation(SearchCriterion.class);
			if (criterion != null) {

				if (field.getType().isPrimitive()) {
					throw new RuntimeException("Primitive types cannot be used as search criteria.  Impossible to tell if a primitive type is excluded from a search.");
				}

				// This field is annotated as a search criterion.  Check to see if the field has a non-null value, and
				// if it doesn't have one, don't include it in the query.
				Object value;
				try {
					value = field.get(searchCriteria);
					if (value == null) {
						// If the criterion contains no value, skip it.
						continue;
					}
					if (value instanceof String && !criterion.useEmptyStringValue() && value.equals("")) {
						// Don't apply an empty string if the annotation says not to.
						continue;
					}
				}
				catch (IllegalAccessException e) {
					throw new RuntimeException("Unable to access field " + searchCriteria.getClass().getName() + "." + field.getName());
				}

				// The operand is a property name.
				String operand = criterion.propertyName();
				if (operand.equals("")) {
					// If no property name specified in annotation, default to field name.
					operand = field.getName();
				}

				// Lazy instantiation of predicate list--only build when search criteria specified.
				if (predicates == null) {
					predicates = new ArrayList<Predicate>();
				}

				if (StringUtils.isNotBlank(criterion.joinPath())) {
					predicates.add(new Predicate(criterion.joinPath(), operand, criterion.operator(), value));
				}
				else {
					predicates.add(new Predicate(operand, criterion.operator(), value));
				}

//				applyCriterion(query, criterion.operator(), value, operand);
			}
		}

		if (predicates != null) {
			applyPredicates(query, predicates);
		}
	}

	/**
	 * Converts a predicate object into the appropriate hibernate Criteria API call.
	 *
	 * @param query
	 * @param operand
	 * @param operator
	 * @param value
	 */
	private void applyPredicate(Criteria query, String operand, Operator operator, Object value)
	{
		switch (operator)
		{
			// Convert the generic DAO operation into a hibernate restriction and add it to the
			// query.
			case EqualTo:
				query.add(Restrictions.eq(operand, value));
				break;
			case GreaterThan:
				query.add(Restrictions.gt(operand, value));
				break;
			case GreaterThanOrEqual:
				query.add(Restrictions.ge(operand, value));
				break;
			case NotEqualTo:
				query.add(Restrictions.ne(operand, value));
				break;
			case IsNotNull:
				query.add(Restrictions.isNotNull(operand));
				break;
			case IsNull:
				query.add(Restrictions.isNull(operand));
				break;
			case IsIn:
				Class operandType = value.getClass();
				if (Collection.class.isAssignableFrom(operandType)) {
					query.add(Restrictions.in(operand, (Collection)value));
				}
				else if (operandType.isArray()) {
					Class arrayType = operandType.getComponentType();
					if (arrayType.isPrimitive()) {
						// primitive array needs to be converted into object array
						int length = Array.getLength(operand);
						Object[] arrayValues = new Object[length];
						for (int i = 0; i < Array.getLength(operand); i++) {
							arrayValues[i] = Array.get(operand, i);
						}
						query.add(Restrictions.in(operand, arrayValues));
					}
					else {
						query.add(Restrictions.in(operand, (Object[])value));
					}
				}
				break;
			case LessThan:
				query.add(Restrictions.lt(operand, value));
				break;
			case LessThanOrEqual:
				query.add(Restrictions.le(operand, value));
				break;
			case Matches:
				query.add(Restrictions.ilike(operand, value.toString(), MatchMode.ANYWHERE));
				break;
			case StartsWith:
				query.add(Restrictions.ilike(operand, value.toString(), MatchMode.START));
				break;
			case EndsWith:
				query.add(Restrictions.ilike(operand, value.toString(), MatchMode.END));
				break;
			case IsLike:
				query.add(Restrictions.ilike(operand, value.toString().replace('*', '%')));
				break;
			case CountEqualTo:
				query.add(Restrictions.sizeEq(operand, ((Number)value).intValue()));
				break;
			case CountGreaterThan:
				query.add(Restrictions.sizeGt(operand, ((Number)value).intValue()));
				break;
			case CountLessThan:
				query.add(Restrictions.sizeLt(operand, ((Number)value).intValue()));
				break;
			case CountNotEqualTo:
				query.add(Restrictions.sizeNe(operand, ((Number)value).intValue()));
				break;
			case CountGreaterThanOrEqual:
				query.add(Restrictions.sizeGe(operand, ((Number)value).intValue()));
				break;
			case CountLessThanOrEqual:
				query.add(Restrictions.sizeLe(operand, ((Number)value).intValue()));
				break;
		}
	}

	public <T extends IEntity> GridDisplayResults<T> search(Class<T> entityClass, SearchCriteria searchCriteria, GridDisplayOptions displayOptions, Enum... fetchOptions)
	{
		return search(entityClass, entityClass, searchCriteria, displayOptions, fetchOptions);
	}

	public <T extends IEntity, Y extends IEntity> GridDisplayResults<T> search(Class<T> viewClass, Class<Y> criteriaClass, SearchCriteria searchCriteria, GridDisplayOptions displayOptions, Enum... fetchOptions)
	{
		GridDisplayResults<T> results = new GridDisplayResults<T>();

		Enum prefetch = null;

		// Don't to a total record count if the display options indicate that all results are to be
		// returned in this query.
		if (displayOptions != null && displayOptions.getPageSize() != GridDisplayOptions.ALL_RESULTS) {

			//TODO: If any collection on this entity is eagerly fetched,

			Criteria countQuery = getSession().createCriteria(criteriaClass);

			applySearchCriteria(searchCriteria, countQuery);

			// Return a count, not the actual entity list.
			countQuery.setProjection(Projections.countDistinct("id"));

			// Save this value on the return object.
			Number count = (Number)countQuery.uniqueResult();
			results.setTotalItemCount(count.intValue());

			// Now run the actual query, applying paging parameters.

			// Check to see if this entity has a composite key.
			boolean entityHasEmbeddedId = false;
			Field idField = FieldUtils.getDeclaredField(viewClass, "id");
			EmbeddedId embedded = (EmbeddedId)idField.getAnnotation(EmbeddedId.class);
			if (embedded != null) {
				entityHasEmbeddedId = true;
			}

			//TODO: find a way to unify these two query approaches into one.  It shouldn't be this difficult to come up
			//TODO: with a distinct set of results for a hibernate query.  Perhaps add all entity fields to a projection
			//TODO: list with a distinct projection wrapper around it?
			if (entityHasEmbeddedId || (! hasMultiValueField(viewClass) && viewClass.equals(criteriaClass))) {
				// This is the less than ideal way to run this query, necessary because the "ideal" way doesn't work
				// with a composite primary key.
				Criteria resultsQuery = getSession().createCriteria(viewClass);

				prefetch = applyPrefetch(viewClass, resultsQuery, fetchOptions);

				applySearchCriteria(searchCriteria, resultsQuery);

				applySorting(viewClass, resultsQuery, displayOptions);

				applyPaging(resultsQuery, displayOptions);

				// DON'T USE THIS API--THE LIST TRIMMING FOR DISTINCT AFTER DB CALL, THROWING OFF PAGING.
				//resultsQuery.setResultTransformer(Criteria.DISTINCT_ROOT_ENTITY);

				results.setItems(resultsQuery.list());
			}
			else {
				// This is the ideal way to run this query, which assumes a single value primary key.
				DetachedCriteria resultsQuery = DetachedCriteria.forClass(criteriaClass);
				Criteria queryAdapter = new DetachedCriteriaAdapter(resultsQuery);
				applySearchCriteria(searchCriteria, queryAdapter);

				resultsQuery.setProjection(
						Projections.distinct(Projections.property("id"))
				);

				Criteria rootQuery = getSession().createCriteria(viewClass);
				prefetch = applyPrefetch(viewClass, rootQuery, fetchOptions);
				rootQuery.add(Subqueries.propertyIn("id", resultsQuery));
				applyPaging(rootQuery, displayOptions);
				applySorting(viewClass, rootQuery, displayOptions);

				results.setItems(rootQuery.list());
			}

			logger.debug("Total items matching criteria: " + results.getTotalItemCount());
		}
		else {
			// The simple scenario--no paging.  Returning all results.  No count query necessary.  Apply the search
			// criteria and the sorting and get all results.
			Criteria resultsQuery = getSession().createCriteria(viewClass);
			prefetch = applyPrefetch(viewClass, resultsQuery, fetchOptions);
			applySearchCriteria(searchCriteria, resultsQuery);
			applySorting(viewClass, resultsQuery, displayOptions);

			results.setItems(resultsQuery.list());

			results.setTotalItemCount(results.getItems().size());
		}

		if (fetchOptions.length > 1) {
			for (T item : results.getItems()) {
				applyPostfetch(viewClass, prefetch, item, fetchOptions);
			}
		}

		return results;
	}

	protected void applyPaging(Criteria resultsQuery, GridDisplayOptions options)
	{
		if (logger.isDebugEnabled() && options != null) {
			logger.debug(MessageFormat.format("Applying paging: index={0}, size={1}", options.getPageIndex(), options.getPageSize()));
		}

		// Skip ahead to page requested.
		resultsQuery.setFirstResult(options.getPageIndex() * options.getPageSize());

		resultsQuery.setMaxResults(options.getPageSize());
	}

	/**
	 * Convenience method for applying order by options to a query.
	 * @param criteria
	 * @param options
	 */
	protected <T extends IEntity> void applySorting(Class<T> entityClass, Criteria criteria, GridDisplayOptions options)
	{
		// Apply sorting by first checking for a parameter value, if not present use a default sort value
		// defined on the entity, otherwise use the id field.
		String sortProperty = null;
		Boolean sortAscending = null;

		// If the options parameter is non-null, attempt to get some sort stuff from it.
		if (options != null) {
			sortProperty = options.getSortProperty();
			sortAscending = options.isAscending();
		}

		// If no sort property was specified, check the class for annotations that specify a default sort property.
		if (sortProperty == null) {
			List<SortSpecification> sorts = DaoUtils.getDefaultSortSpecifications(entityClass);
			if (sorts != null) {
				for (SortSpecification sort : sorts) {

					if (logger.isDebugEnabled()) {
						logger.debug(MessageFormat.format("Applying default sort: property={0}, ascending={1}", sort.getProperty(), sort.isAscending()));
					}

					if (sort.isAscending()) {
						criteria.addOrder(Order.asc(sort.getProperty()));
					}
					else {
						criteria.addOrder(Order.desc(sort.getProperty()));
					}
				}
			}
		}
		else {
			//TODO: this might be an expensive operation for each search.  Consider optimizing via caching.
			// The sort property in use might not be the actual name of a field on this entity class.  It is possible
			// that the property maps to a @SortProperty annotation on one of the fields in this class.
			for (Field field : FieldUtils.getDeclaredFields(entityClass)) {
				if (field.getName().equals(sortProperty)) {
					// Simple case--sort property maps directly to an entity field.
					break;
				}

				SortProperty sortAnnot = field.getAnnotation(SortProperty.class);
				if (sortAnnot != null && sortAnnot.value().equals(sortProperty)) {
					// The sort expression provided matches the sort annotation on this field.  Substitute field name.
					sortProperty = field.getName();
					break;
				}

				EmbeddedId embeddedId = field.getAnnotation(EmbeddedId.class);
				if (embeddedId != null) {
					Field[] embeddedFields = field.getType().getDeclaredFields();
					for (Field embeddedField : embeddedFields) {
						embeddedField.setAccessible(true);

						sortAnnot = embeddedField.getAnnotation(SortProperty.class);
						if (sortAnnot != null && sortAnnot.value().equals(sortProperty)) {
							// Found a sort property match on an embedded field.
							sortProperty = field.getName() + "." + embeddedField.getName();
							break;
						}
					}
				}
			}

			if (logger.isDebugEnabled() && options != null) {
				logger.debug(MessageFormat.format("Applying sorting: property={0}, ascending={1}", sortProperty, sortAscending));
			}

			// If no sort direction was specified, look for a default value.
			if (sortAscending == null) {
				sortAscending = DaoUtils.getDefaultSortAscending(entityClass);
			}

			// Apply the sort property and direction to the query.
			if (sortAscending) {
				criteria.addOrder(Order.asc(sortProperty));
			}
			else {
				criteria.addOrder(Order.desc(sortProperty));
			}			
		}
	}

	public <T extends IEntity> Serializable insert(T entity)
	{
		getSession().save(entity);

		return entity.getId();
	}

	public <T extends IEntity> void saveOrUpdate(T entity)
	{
		getSession().saveOrUpdate(entity);
	}

	public <T extends IEntity> void update(T entity)
	{
		getSession().update(entity);
	}

	public <T extends IEntity> void delete(T entity)
	{
		getSession().delete(entity);
	}

	public <T extends IEntity> void deleteById(Class<T> entityClass, Serializable id)
	{
		StringBuilder deleteSql = new StringBuilder();
		deleteSql.append("delete from ");
		deleteSql.append(entityClass.getName());
		deleteSql.append(" where id = :id");

		Query delete = getSession().createQuery(deleteSql.toString());
		delete.setParameter("id", id);

		int count = delete.executeUpdate();

		if (count == 0) {
			throw new DataNotFoundException(entityClass, id);
		}
	}

	public <T extends IEntity> void deleteAll(Collection<T> entities)
	{
		if (entities.size() == 0)
			return;

		Class entityClass = null;

		Serializable[] entityIds = new Serializable[entities.size()];
		int i = 0;
		for (T entity : entities) {
			if (entityClass == null) {
				entityClass = entity.getClass();
			}
			entityIds[i++] = entity.getId();
		}

		StringBuilder deleteSql = new StringBuilder();
		deleteSql.append("delete from ");
		deleteSql.append(entityClass.getName());
		deleteSql.append(" where id in ( :ids )");

		Query delete = getSession().createQuery(deleteSql.toString());
		delete.setParameterList("ids", entityIds);

		int count = delete.executeUpdate();

		if (count != entities.size()) {
			throw new DataNotFoundException(entityClass, null);
		}
	}

	public Long getCounterValue(String counterName)
	{
		String schemaName = getDefaultSchemaName();
		if (! StringUtils.isBlank(schemaName)) {
			counterName = schemaName + "." + counterName;
		}
		
		BigDecimal dec = (BigDecimal)getSession().createSQLQuery("select " + counterName + ".nextval from dual").uniqueResult();
		return dec.longValue();
	}

	public void saveChanges()
	{
		getSession().flush();
	}

	public <T extends IEntity> List<T> executeQuery(String query, Object params)
	{
		Query hql = getSession().createQuery(query);

		if (params != null) {
			hql.setProperties(params);
		}

		return hql.list();
	}

	public <T extends IEntity> List<T> executeNamedQuery(String name, Object params)
	{
		Query hql = getSession().getNamedQuery(name);

		if (params != null) {
			hql.setProperties(params);
		}

		return hql.list();
	}

	public List<Map<String, ?>> executeSQL(String sql, Map<String, Class<?>> returnValues, Object params)
	{
		SQLQuery query = getSession().createSQLQuery(sql);

		// Tell hibernate what the return values and their types are.
		for (String key : returnValues.keySet()) {
			Type type = getTypeForClass(returnValues.get(key));
			query.addScalar(key, type);
		}

		if (params != null) {
			query.setProperties(params);
		}

		// Use a sorting map to determine the ordering of aliases in the source SQL's select statement.
		TreeMap<Integer, String> aliasMap = new TreeMap<Integer, String>();
		for (String alias : returnValues.keySet()) {
			int pos = sql.indexOf(alias);
			if (pos == -1) {
				throw new RuntimeException("Alias '" + alias + "' not found in provided SQL: " + sql);
			}

			aliasMap.put(pos, alias);
		}

		// This list of aliases will be used to convert the object array based return values into mapped values.
		ArrayList<String> orderedAliases = new ArrayList<String>(aliasMap.values());

		ArrayList<Map<String, ?>> results = new ArrayList<Map<String, ?>>();

		for (Object result : query.list()) {

			Object[] rawResult = (Object[])result;

			HashMap<String, Object> mappedResults = new HashMap<String, Object>();

			int index = 0;
			for (String alias : orderedAliases) {
				mappedResults.put(alias, rawResult[index++]);
			}

			results.add(mappedResults);
		}

		return results;
	}

	public <T extends IEntity> List<T> executeSQL(Class<T> entityClass, String sql, Object params)
	{
		SQLQuery query = getSession().createSQLQuery(sql);

		query.addEntity(entityClass);

		if (params != null) {
			query.setProperties(params);
		}

		return query.list();
	}

	public List<Map<String, ?>> executeNamedSQL(String name, Object params)
	{
		Query hql = getSession().getNamedQuery(name);

		if (params != null) {
			hql.setProperties(params);
		}

		// Use the underlying hibernate session implementation to get access to the definition of the named query.
		// This will provide us with the list return value aliases that will be used to build a Map of each return
		// value.
		NamedSQLQueryDefinition queryDef = ((SessionFactoryImpl)sessionFactory).getNamedSQLQuery(name);
		NativeSQLQueryReturn[] returns = queryDef.getQueryReturns();

		// This list of aliases will be used to convert the object array based return values into mapped values.
		ArrayList<String> orderedAliases = new ArrayList<String>();
		for (NativeSQLQueryReturn queryReturn : returns) {
			if (! (queryReturn instanceof NativeSQLQueryScalarReturn)) {
				throw new RuntimeException("Only scalar return types supported. Why are you getting all fancy with your named SQL?");
			}
			orderedAliases.add(((NativeSQLQueryScalarReturn)queryReturn).getColumnAlias());
		}

		ArrayList<Map<String, ?>> results = new ArrayList<Map<String, ?>>();

		for (Object result : hql.list()) {
			Object[] rawResults = (Object[])result;

			HashMap<String, Object> mappedResults = new HashMap<String, Object>();

			int index = 0;
			for (String alias : orderedAliases) {
				mappedResults.put(alias, rawResults[index++]);
			}

			results.add(mappedResults);
		}

		return results;
	}

	public <T extends IEntity> List<T> executeNamedEntitySQL(String name, Object params)
	{
		Query hql = getSession().getNamedQuery(name);

		if (params != null) {
			hql.setProperties(params);
		}

		return hql.list();
	}
}
