/*
 * Copyright 2010 Portland Webworks, Inc.
 *
 * This file is part of PWW Commons.
 *
 * PWW Commons is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 * Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * PWW Commons is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with PWW Commons.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

package com.portlandwebworks.commons.hibernate;

import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;
import org.hibernate.MappingException;
import org.hibernate.dialect.Dialect;
import org.hibernate.id.PersistentIdentifierGenerator;
import org.hibernate.id.SequenceGenerator;
import org.hibernate.type.Type;

import java.util.Properties;

/**
 * This software is the property of Portland Webworks.
 *
 * This custom sequence generator is tailored to Oracle sequences.  It assumes that the name of the
 * sequence used to generate primary key values is the table name prefixed with SEQ_ so that one doesn't
 * have to name the sequence in the hibernate config.
 *
 *
 * @Author rnorthup
 */
public class TableNameSequenceGenerator extends SequenceGenerator
{
	public static final int MAX_SEQ_NAME_LENGTH = 30;

	private Logger logger = Logger.getLogger(TableNameSequenceGenerator.class);

    /**
     * If the parameters do not contain a {@link org.hibernate.id.SequenceGenerator#SEQUENCE} name, we
     * assign one based on the table name.  The naming convention is to prefix the table name with
	 * SEQ_ and to ensure that the sequence name is no longer than
     */
    public void configure(Type type, Properties params, Dialect dialect) throws MappingException
	{
		//todo: this logic doesn't allow assiging a sequence to an entity field other than the primary
		//todo: key of the entity--it will fabricate a sequence name that is a duplicate.
        if (StringUtils.isBlank(params.getProperty(SEQUENCE))) {
            String tableName = params.getProperty(PersistentIdentifierGenerator.TABLE);
            if(tableName != null) {
                String seqName = StringUtils.substring("SEQ_" + tableName, 0, MAX_SEQ_NAME_LENGTH);

				logger.info("Generated default sequence name for entity " + type.getName() + ": " + seqName);

                params.setProperty(SEQUENCE, seqName);
            }
        }
		else {
			logger.warn("No " + SEQUENCE + " property defined for type " + type.getName() + ", unable to generate a default sequence name");
		}
        super.configure(type, params, dialect);
    }
}