/*
 * Copyright 2010 Portland Webworks, Inc.
 *
 * This file is part of PWW Commons.
 *
 * PWW Commons is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 * Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * PWW Commons is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with PWW Commons.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

package com.portlandwebworks.commons.servlet;
import com.lowagie.text.DocumentException;
import com.portlandwebworks.commons.html.HtmlUtils;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;
import org.springframework.mock.web.DelegatingServletOutputStream;
import org.w3c.dom.Document;
import org.w3c.tidy.Tidy;
import org.xhtmlrenderer.pdf.ITextRenderer;

import javax.servlet.*;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpServletResponseWrapper;
import java.io.*;

/**
 * This software is the property of Portland Webworks.
 *
 * @Author rnorthup
 */
public class PdfFilter implements Filter
{
	private Logger logger = Logger.getLogger(PdfFilter.class);
	protected ITextRenderer renderer;

	public void init(FilterConfig filterConfig) throws ServletException
	{
		renderer = new ITextRenderer();
	}

	public void doFilter(ServletRequest servletRequest, ServletResponse servletResponse, FilterChain filterChain) throws IOException, ServletException
	{
		HttpServletRequest request = (HttpServletRequest)servletRequest;
		HttpServletResponse response = (HttpServletResponse)servletResponse;

		// Pass the request along in the filter chain--but include with it a mock response object that will capture
		// the rendered HTML so we can process it.
		CapturedServletResponse capturedResponse = new CapturedServletResponse(response);

		filterChain.doFilter(request, capturedResponse);

		String content = capturedResponse.toString();

		// Construct the base URL to be inerted before all relative URLs in the document.
		String url = request.getRequestURL().toString();
		String context = request.getContextPath();
		String protocolHostAndPort = StringUtils.substringBefore(url, context);
		content = StringUtils.replace(content, context, protocolHostAndPort + context);

		//content = content.replaceAll("<thead>|</thead>|" + "<tbody>|</tbody>", "");

		// Use the title of the document as the attachment filename
		String title = HtmlUtils.extractTitleFromHtml(content);
		if (title == null) {
			// XHTML-compliant document must have a title, so this means our title extraction code is failing.
			throw new RuntimeException("HTML document title extraction failed! \n" + content);
		}

		response.setContentType("application/pdf");

		// Indicate to the requesting client that the response is a file attachment
		response.setHeader("content-disposition", "attachment;filename=\"" + title + ".pdf\"");

		OutputStream out = response.getOutputStream();

		try {
			Tidy tidy = new Tidy();

			Document document = tidy.parseDOM(new StringReader(content), null);

			// Tell iText to render the PDF to the "real" servlet output stream using the temp file
			renderer.setDocument(document, document.getBaseURI());
			renderer.layout();
			renderer.createPDF(out);
			out.flush();
		}
		catch (Exception e) {
			logger.error("Unhandled exception generating PDF for HTML content", e);
		}
		finally {
			IOUtils.closeQuietly(out);
		}
	}

	public void destroy()
	{
		//To change body of implemented methods use File | Settings | File Templates.
	}
}

class CapturedServletResponse extends HttpServletResponseWrapper
{
	private ByteArrayOutputStream baos;

	private PrintWriter writer;

	private ServletOutputStream stream;

	public String toString()
	{
		return baos.toString();
	}

	public CapturedServletResponse(HttpServletResponse response)
	{
		super(response);

		baos = new ByteArrayOutputStream();
	}

	@Override
	public ServletOutputStream getOutputStream() throws IOException
	{
		if (writer != null) {
			throw new RuntimeException("Writer already obtained from response, can't create an output stream too");
		}

		if (stream == null) {
			stream = new DelegatingServletOutputStream(baos);
		}

		return stream;
	}

	public PrintWriter getWriter()
	{
		if (stream != null) {
			throw new RuntimeException("Output stream already obtained from response, can't create a writer too");
		}

		if (writer == null) {
			writer = new PrintWriter(baos);
		}

		return writer;
	}

}
