/*
 * Copyright 2010 Portland Webworks, Inc.
 *
 * This file is part of PWW Commons.
 *
 * PWW Commons is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 * Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * PWW Commons is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with PWW Commons.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

package com.portlandwebworks.commons.test.dbunit;

import org.apache.log4j.Logger;
import org.dbunit.DatabaseUnitException;
import org.dbunit.database.*;
import org.dbunit.dataset.datatype.DefaultDataTypeFactory;
import org.dbunit.ext.db2.Db2DataTypeFactory;
import org.dbunit.ext.db2.Db2MetadataHandler;
import org.dbunit.ext.h2.H2DataTypeFactory;
import org.dbunit.ext.hsqldb.HsqldbDataTypeFactory;
import org.dbunit.ext.mssql.MsSqlDataTypeFactory;
import org.dbunit.ext.mysql.MySqlDataTypeFactory;
import org.dbunit.ext.mysql.MySqlMetadataHandler;
import org.dbunit.ext.oracle.Oracle10DataTypeFactory;
import org.dbunit.ext.oracle.OracleDataTypeFactory;
import org.dbunit.ext.postgresql.PostgresqlDataTypeFactory;
import org.hibernate.Session;
import org.hibernate.SessionFactory;
import org.hibernate.dialect.*;
import org.hibernate.impl.SessionFactoryImpl;
import org.springframework.beans.factory.annotation.Autowired;

import java.sql.SQLException;

/**
 * This software is the property of Portland Webworks.
 *
 * @Author rnorthup
 */
public class HibernateConnectionProvider implements IDatabaseConnectionProvider
{
	/**
	 * Have Spring provide the session factory as defined in the spring context.
	 */
	@Autowired(required = true)
	SessionFactory sessionFactory;

	private Logger logger = Logger.getLogger(HibernateConnectionProvider.class);

	private ThreadLocal<DatabaseConnection> currentConnection = new ThreadLocal<DatabaseConnection>();

	private ThreadLocal<Session> currentSession = new ThreadLocal<Session>();

	public IDatabaseConnection setupDatabaseConnection()
	{
		// Get the schema name using some hibernate hackery.
		String schema;
		Class<? extends DefaultDataTypeFactory> dataTypeFactory;
		Class<? extends IMetadataHandler> metaDataHandler = DefaultMetadataHandler.class;
		if (sessionFactory instanceof SessionFactoryImpl) {
			SessionFactoryImpl impl = (SessionFactoryImpl)sessionFactory;

			schema = impl.getSettings().getDefaultSchemaName();
			logger.info("Applying DbUnit dataset to schema " + schema);

			dataTypeFactory = getDbUnitDataTypeFactoryForHibernateDialect(impl.getSettings().getDialect());
			logger.info("Using dataTypeFactory " + dataTypeFactory.getName());

			if (MySqlDataTypeFactory.class.isAssignableFrom(dataTypeFactory)) {
				metaDataHandler = MySqlMetadataHandler.class;
			}
			else if (Db2DataTypeFactory.class.isAssignableFrom(dataTypeFactory)) {
				metaDataHandler = Db2MetadataHandler.class;
			}
		}
		else {
			throw new RuntimeException("SessionFactory is not of a known subclass: " + sessionFactory.getClass().getName());
		}

		// Start a hibernate session
		logger.info("Opening hibernate session...");
		Session session = sessionFactory.openSession();

		// Why doesn't hibernate put this in thread local storage for me?
		currentSession.set(session);

		try {
			currentConnection.set(new DatabaseConnection(session.connection(), schema));
		}
		catch (DatabaseUnitException e) {
			throw new RuntimeException("Unable to initialize database connection", e);
		}

		//TODO: only set this property if we are running on Oracle 10g or later.
		if (Oracle10DataTypeFactory.class.equals(dataTypeFactory)) {
			currentConnection.get().getConfig().setProperty(DatabaseConfig.FEATURE_SKIP_ORACLE_RECYCLEBIN_TABLES, true);
		}

		try {
			currentConnection.get().getConfig().setProperty(DatabaseConfig.PROPERTY_DATATYPE_FACTORY, dataTypeFactory.newInstance());
		}
		catch (Exception e) {
			logger.error("Unhandled exception setting datatype factory " + dataTypeFactory.getName(), e);
		}

		try {
			currentConnection.get().getConfig().setProperty(DatabaseConfig.PROPERTY_METADATA_HANDLER, metaDataHandler.newInstance());
		}
		catch (Exception e) {
			logger.error("Unhandled exception setting metadata handler " + metaDataHandler.getName(), e);
		}

		return currentConnection.get();
	}

	public void tearDownDatabaseConnection()
	{
		Session session = currentSession.get();
		if (session != null && session.isOpen()) {
			session.close();
		}

		try {
			if (currentConnection.get() != null) {
				currentConnection.get().close();
			}
		}
		catch (SQLException e) {
			logger.error("Error closing DbUnit connection", e);
		}

		currentConnection.remove();
	}

	/**
	 * Translates a Hibernate SQL dialect class to the appropriate DbUnit data type factory class.
	 *
	 * @param dialect
	 * @return
	 */
	private Class<? extends DefaultDataTypeFactory> getDbUnitDataTypeFactoryForHibernateDialect(Dialect dialect)
	{
		Class<? extends DefaultDataTypeFactory> dataTypeFactory = DefaultDataTypeFactory.class;

		Class<? extends Dialect> dialectClass = dialect.getClass();
		if (Oracle10gDialect.class.isAssignableFrom(dialectClass)) {
			dataTypeFactory = Oracle10DataTypeFactory.class;
		}
		else if (OracleDialect.class.isAssignableFrom(dialectClass)) {
			dataTypeFactory = OracleDataTypeFactory.class;
		}
		else if (DB2Dialect.class.isAssignableFrom(dialectClass)) {
			dataTypeFactory = Db2DataTypeFactory.class;
		}
		else if (MySQLDialect.class.isAssignableFrom(dialectClass)) {
			dataTypeFactory = MySqlDataTypeFactory.class;
		}
		else if (HSQLDialect.class.isAssignableFrom(dialectClass)) {
			dataTypeFactory = HsqldbDataTypeFactory.class;
		}
		else if (H2Dialect.class.isAssignableFrom(dialectClass)) {
			dataTypeFactory = H2DataTypeFactory.class;
		}
		else if (SQLServerDialect.class.isAssignableFrom(dialectClass)) {
			dataTypeFactory = MsSqlDataTypeFactory.class;
		}
		else if (PostgreSQLDialect.class.isAssignableFrom(dialectClass)) {
			dataTypeFactory = PostgresqlDataTypeFactory.class;
		}

		return dataTypeFactory;
	}
}
